// ============================================================================
// ============================================================================
// ============================================================================
// ==                                                                        ==
// == Name    : TheEmuLib.Emu_Ripple_XY.fsh                                  ==
// == Type    : Fragment shader                                              ==
// == Version : 1.2.14 (2017/01/06)                                          ==
// == Creator : TheEmu © TheEmu 2017, Some Rights Reserved                   ==                                                   ==
// == Licence : Creative Commons Attribution-ShareAlike 4.0                  ==
// ==           http://creativecommons.org/licences/by-sa/4.0                ==
// ==                                                                        ==
// == Purpose: To apply a "ripple" to the (x,y) coordinates of the points in ==
// == an image with the ripple being a function of their (x,y) coordinates.  ==
// ==                                                                        ==
// == Description:  A  source image is modified by applying a time dependant ==
// == wobbling motion. The ripple at any point comprises a sum of up to four ==
// == terms added to its normalised coordinates, each term being of the form ==
// ==                                                                        ==
// ==        K0 * Ax * cos ( T*Fx - B*x - C*y + Dx ) + K1                    ==
// ==        K0 * Ay * cos ( T*Fy - B*y - C*x + Dy ) + K1                    ==
// ==                                                                        ==
// == where T is time, (x,y) the normalised coordinates and A, B, C, D, F, K ==
// == are ripple parameters supplied as shader inputs.  The negation of some ==
// == terms involving B, C, D and K is done so that positive values of these ==
// == parameters correspond to movement in a positive direction.             ==
// ==                                                                        ==
// == Static distortion of the source image is obtained using a frequency of ==
// == of 0.0 for all components of the ripple.                               ==
// ==                                                                        ==
// == This file is a member of The Emu's shader library.                     ==
// ==                                                                        ==
// == ====================================================================== ==
// ==                                                                        ==
// == WARNING: Currently this shader does not work as I want when very large ==
// == ripple amplitudes are used. It is OK for total amplitudes less than or ==
// == equal to 1.0, but for larger amplitudes the effect is not what I want. ==
// == As a ripple amplitude of 1.0 means that rippling motion is much larger ==
// == than the window dimensions the problem, if it is real, will not affect ==
// == normal use of the shader. However, be aware that I hope to correct the ==
// == shader's behaviour even for any extreme case so you should not rely on ==
// == the shader for such large ripple amplitudes. The behaviour for ripples ==
// == with an amplitude less than 1.0 can howver be relied upon. TheEmu.     ==
// ==                                                                        ==
// == ====================================================================== ==
// ==                                                                        ==
// == Update history:                                                        ==
// ==                                                                        ==
// ==   2016/12/07 - v1.0.0 - Initial version.                               ==
// ==   2016/12/10 - v1.1.0 - Added edge mode control.                       ==
// ==   2016/12/12 - v1.1.1 - Cleaned up a little.                           ==
// ==   2016/12/14 - v1.2.0 - Added border option.                           ==
// ==   2016/12/15 - v1.2.1 - Minor clean up.                                ==
// ==   2016/12/16 - v1.2.2 - Regularised format of TheEmuLib sources.       ==
// ==   2016/12/17 - v1.2.3 - Phases now in cycles for consistancy.          ==
// ==   2016/12/17 - v1.2.4 - Borders replaced by edge coordinates.          ==
// ==   2016/12/18 - v1.2.5 - Added an option to drop the unrippled region.  ==
// ==   2016/12/19 - v1.2.6 - Correct error for sign of Cy.                  ==
// ==   2016/12/20 - v1.2.7 - Added overall ripple scale and offset.         ==
// ==   2016/12/21 - v1.2.8 - Changed signs of B and C terms.                ==
// ==   2017/01/05 - v1.2.9 - Renamed all Wobble shaders to be Ripple.       ==
// ==   2017/01/05 - v1.2.10 - Bug fix for edge_hold logic.                  ==
// ==   2017/01/05 - v1.2.11 - Increased commonality with Ripple_RA.         ==
// ==   2017/01/06 - v1.2.12 - mix(vec4,vec4,bvec4) not allowed by NVIDIA.   ==
// ==   2017/01/06 - v1.2.13 - Interchanged edge_modes 1 and 2.              ==
// ==   2017/01/06 - v1.2.14 - Fixed bug in edge smearing logic.             ==
// ==                                                                        ==
// ============================================================================
// ============================================================================
// ============================================================================

// ============================================================================
// == Standard shader inputs ==================================================
// ============================================================================

uniform float u_Elapsed;    // The elapsed time in seconds
uniform vec2  u_WindowSize; // Window dimensions in pixels

// The image that is to be manipulated.

uniform sampler2D iChannel0;

// ============================================================================
// == Imports from TheEmuLib ==================================================
// ============================================================================
//
// The GLSL shader language currently provides no mechanism for importing  any
// elements that are defined in other modules, not even C's crude source level
// #include mechanism. In the absence of anything better TheEmuLib handles any
// imports by manualy copying relevent utility code snippets from the  sources
// in the Shader Lib.Inc directory. This is very crude but I have attempted to
// be systematic in the way in which this is presented in the library sources.
//
// = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =

// Macros from TheEmuLib.Emu_Basic_Constants.lib.src

#define PI  3.1415926535897932384626433832795
#define TAU 6.2831853071795864769252867665590

// = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =

// Macros from TheEmuLib.Emu_Common_Utilities.lib.src

#define EMU_DEFAULT(type,x,default_value) (x==type(0.0)) ? (default_value) : (x)

// = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =

// Functions from TheEmuLib.Emu_Coordinate_Normalisation.lib.src

#define EMU_NORMALISE_TO_WINDOW_1(xy,wsize) (     (xy)/(wsize) )
#define EMU_NORMALISE_TO_WINDOW_2(xy,wsize) ( 2.0*(xy)/(wsize) - 1.0 )

#define EMU_NORMALISE_TO_WINDOW_3(xy,c1,c2) ( EMU_NORMALISE_TO_WINDOW_1(xy-c1,c2-c1) )
#define EMU_NORMALISE_TO_WINDOW_4(xy,c1,c2) ( EMU_NORMALISE_TO_WINDOW_2(xy-c1,c2-c1) )

#define EMU_DENORMALISE_TO_WINDOW_1(uv,size)  (      (uv)        * (size) )
#define EMU_DENORMALISE_TO_WINDOW_2(uv,size)  ( ( (uv)*0.5+0.5 ) * (size) )

#define EMU_DENORMALISE_TO_WINDOW_3(uv,c1,c2) ( EMU_DENORMALISE_TO_WINDOW_1(uv,c2-c1)+c1 )
#define EMU_DENORMALISE_TO_WINDOW_4(uv,c1,c2) ( EMU_DENORMALISE_TO_WINDOW_2(uv,c2-c1)+c1 )

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

vec2 Emu_Normalise_to_Window ( vec2 xy )
 { return EMU_NORMALISE_TO_WINDOW_1 ( xy, u_WindowSize.xy );
 }

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

vec2 Emu_Normalise_to_Window ( vec2 xy, vec2 xy0, vec2 xy1 ) 
 { return EMU_NORMALISE_TO_WINDOW_3 ( xy, xy0, xy1 );
 }

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

vec2 Emu_Denormalise_to_Window ( vec2 uv, vec2 xy0, vec2 xy1 ) 
 { return EMU_DENORMALISE_TO_WINDOW_3 ( uv, xy0, xy1 );
 }

// = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =

// Macros from TheEmuLib.Emu_Boolean_Vector_Ops.lib.src

#define Emu_bvec4_or(A,B)  bvec4 ( A[0]||B[0], A[1]||B[1], A[2]||B[2], A[3]||B[3] )
#define Emu_bvec4_and(A,B) bvec4 ( A[0]&&B[0], A[1]&&B[1], A[2]&&B[2], A[3]&&B[3] )
#define Emu_bvec4_xor(A,B) bvec4 ( A[0]^^B[0], A[1]^^B[1], A[2]^^B[2], A[3]^^B[3] )

// ============================================================================
// == Shader specific inputs ==================================================
// ============================================================================

// By default the whole area of the source image is rippled, but this can be
// overridden by using the following parameter to limit the to wobbling to a
// particular region.  The  edges parameter defines the region that is to be
// rippled by specifying the low and high limits of normalised coordinates. 

uniform vec4 Emu_Ripple_XY_edges;

vec4 edges = 
  EMU_DEFAULT ( vec4, Emu_Ripple_XY_edges,vec4(0.0,1.0,0.0,1.0) );

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

// The effect of the wobbling at the edges of the rippled area can result in
// discontinuities in the generated image.  By default any tears between the
// rippled area and its suroundings represented by fully transparent  black,
// i.e. they become holes through which any background image can be seen.
//
// The default behaviour of producing tears at the edges can be overidden by
// using the edge_mode parameter, which may have the values:-
//
//   0 - edge_tears - Default, produces "tears" at the edge
//   1 - edge_hold1 - Ripples at edge and orthogonal to it held to 0.0
//   2 - edge_hold2 - Ripples at edge and parallel to it held to 0.0
//   3 - edge_hold3 - Combination of edge_hold1 and edge_hold2
//   4 - edge_wraps - Ripple wraps round to the opposite side 
//   5 - edge_smear - Any tear is filled with the edge's colour
//   6 - edge_selfs - The source image itself fills edge tears
//
// In mode 0 any background image is seen through a tear.
// In mode 1 ripples parallel to the edge fade out near the edge
// In mode 2 ripples orthogonal to the edge fade out near the edge
// In mode 3 ripples of either type fade out near the edge
// In mode 4 ripples from opposing edges wrap round, eliminating tears
// In mode 5 the edge, in effect, is smeared though a tear.
// In mode 6 the source image acts as a background in a tear
//
// If 10 is added to the edge mode then anything outside the rippled area is
// rendered transparent rather than retaining its original colour.
//
// The edge mode is specified separately for all four edges using  the  vec4
// input parameter Emu_Ripple_XY_edge_mode where the elements give the modes
// for the edges at X=0, X=1, Y=0 and Y=1 in that order.  Unfortunately  the
// iStripper program does not support passing an ivec4 from the  scene  file
// so we have to use a vec4 and ignore any fractional part.

uniform vec4 Emu_Ripple_XY_edge_mode;

ivec4 edge_mode1 = ivec4 ( mod(Emu_Ripple_XY_edge_mode/10.0,10.0) );
ivec4 edge_mode2 = ivec4 ( mod(Emu_Ripple_XY_edge_mode,10.0) );

bvec4 edge_drops = equal(edge_mode1,ivec4(1));

bvec4 edge_undef = greaterThan ( edge_mode2, vec4(6) );

bvec4 edge_selfs = equal(edge_mode2,ivec4(6));
bvec4 edge_smear = equal(edge_mode2,ivec4(5));
bvec4 edge_wraps = equal(edge_mode2,ivec4(4));
bvec4 edge_hold3 = equal(edge_mode2,ivec4(3));

bvec4 edge_hold2 = Emu_bvec4_or ( equal(edge_mode2,ivec4(2)), edge_hold3 );
bvec4 edge_hold1 = Emu_bvec4_or ( equal(edge_mode2,ivec4(1)), edge_hold3 );
bvec4 edge_tears = Emu_bvec4_or ( equal(edge_mode2,ivec4(0)), edge_undef );

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

// Overall scale and offset for the ripple, the K0 and K1 parameters refered
// to in the initial description of the shader. K0 is not strictly necessary
// as the same effect could be obtained in the scene file by scaling each of
// the A terms, but it is convenient for the shader user to be able to tweak
// the overall scale rather than changing all of the A parameters.  K1  just
// adds a constant offset to the ripple,  the  same effect could be obtained
// in the scene file by altering the position of the object being rippled. 

uniform vec2 Emu_Ripple_XY_KK;  // Ripple overall scale and offset

vec2 KK = EMU_DEFAULT ( vec2, Emu_Ripple_XY_KK, vec2(1.0,0.0) );

// The ripple parameters for each component of the ripple. Any parameter not
// specified defaults to zero. The spatial frequency terms are the number of
// ripples per unit distance.

uniform vec4 Emu_Ripple_XY_Ax;  // Amplitudes
uniform vec4 Emu_Ripple_XY_Ay;

uniform vec4 Emu_Ripple_XY_Bx;  // Spatial frequencies (X component terms)
uniform vec4 Emu_Ripple_XY_By;

uniform vec4 Emu_Ripple_XY_Cx;  // Spatial frequencies (Y component terms)
uniform vec4 Emu_Ripple_XY_Cy;

uniform vec4 Emu_Ripple_XY_Dx;  // Phases in fractions of a cycle
uniform vec4 Emu_Ripple_XY_Dy;

uniform vec4 Emu_Ripple_XY_Fx;  // Frequencies in Hz
uniform vec4 Emu_Ripple_XY_Fy;

// Ripple parameters as used in the body of the shader.

#define Ax ( Emu_Ripple_XY_Ax )
#define Ay ( Emu_Ripple_XY_Ay )

#define Bx ( Emu_Ripple_XY_Bx * TAU )
#define By ( Emu_Ripple_XY_By * TAU )

#define Cx ( Emu_Ripple_XY_Cx * TAU )
#define Cy ( Emu_Ripple_XY_Cy * TAU )

#define Dx ( Emu_Ripple_XY_Dx * TAU )
#define Dy ( Emu_Ripple_XY_Dy * TAU )

#define Fx ( Emu_Ripple_XY_Fx * TAU )
#define Fy ( Emu_Ripple_XY_Fy * TAU )

// ============================================================================
// == The shader's major function =============================================
// ============================================================================

vec4 Emu_Ripple_XY ( vec2 uv0 )
 {
   // Normalise to the region that is to be rippled.

   vec2 pq = Emu_Normalise_to_Window ( uv0, edges.xz, edges.yw );

   // If outside the rippled area then no ripple is applied and we just have
   // to return the colour corresponding to the unrippled coordinates unless
   // the outer region is to be dropped, i.e. rendered transparent.

   if ( any ( greaterThan ( abs(pq-0.5), vec2(0.5) ) ) ) 
    { // mix(vec4,vec4,bvec4) not recognised by NVIDIA so convert edge_drops.
      return mix ( texture2D(iChannel0,uv0), vec4(0.0), vec4(edge_drops) );
    }

   // Determine which edges affect the ripple at this point.

   bvec4 edge = bvec4 ( pq.x < 0.5, pq.x > 0.5,
                        pq.y < 0.5, pq.y > 0.5
                      );

   // Update the edge flags to reflect the nearest edges.

   edge_tears = Emu_bvec4_and ( edge, edge_tears );
   edge_hold1 = Emu_bvec4_and ( edge, edge_hold1 );
   edge_hold2 = Emu_bvec4_and ( edge, edge_hold2 );
   edge_wraps = Emu_bvec4_and ( edge, edge_wraps );
   edge_smear = Emu_bvec4_and ( edge, edge_smear );
   edge_selfs = Emu_bvec4_and ( edge, edge_selfs );

   // Calculate the four components for both the x and y ripples.
   // See comments at top of shader for why B and C are negated.

   vec4 w1 = Ax * cos ( u_Elapsed*Fx - Bx*pq.x - Cx*pq.y + Dx );
   vec4 w2 = Ay * cos ( u_Elapsed*Fy - By*pq.y - Cy*pq.x + Dy );

   // Apply overall ripple scale and offset parameters.

   w1 = KK[0]*w1 + KK[1];
   w2 = KK[0]*w2 + KK[1];

   // Sum the four components of the wobbling for each axis.

   vec2 ww = vec2 ( dot ( vec4(1.0), w1 ),
                    dot ( vec4(1.0), w2 )
                  );

   // Optionaly force the ripples to to be zero at the edges. This is done
   // by multiplying them by a function, sin, that is zero at the edges.
   //
   // Ripples orthogonal to the edges are controlled by edge_hold1.
   // Ripples parallel to the edges are controlled by edge_hold2.

   vec2 ff = sin(pq*PI);

   if ( edge_hold1[0] || edge_hold1[1] ) ww.x *= ff.x;
   if ( edge_hold1[2] || edge_hold1[3] ) ww.y *= ff.y;

   if ( edge_hold2[0] || edge_hold2[1] ) ww.y *= ff.x;
   if ( edge_hold2[2] || edge_hold2[3] ) ww.x *= ff.y;

   // Apply the ripple to the normalised coordinates.

   pq += ww;

   // Determine if there is to be a tear, and if so if the source image is
   // to be seen through the tear (edge_selfs mode) or if there is to be a
   // a hole through which any background is visible (edge_tears mode).

   bvec2 tear = bvec2 ( pq.x != fract(pq.x), pq.y != fract(pq.y) );

   edge_tears = Emu_bvec4_and ( edge_tears, tear.xxyy );
   edge_selfs = Emu_bvec4_and ( edge_selfs, tear.xxyy );

   // Perform any edge wrap around where it is enabled.

   if ( edge_wraps[0] || edge_wraps[1] ) pq.x = fract(pq.x);
   if ( edge_wraps[2] || edge_wraps[3] ) pq.y = fract(pq.y);

   // Apply edge smearing where it is enabled.

   if ( edge_smear[0] || edge_smear[1] ) clamp ( pq.x, 0.0, 1.0 );
   if ( edge_smear[2] || edge_smear[3] ) clamp ( pq.y, 0.0, 1.0 );

   // Renormalise to the dimensions of the whole image.

   vec2 st = Emu_Denormalise_to_Window ( pq, edges.xz, edges.yw );

   // Optionaly use original coordinates for holes.

   if ( any ( edge_selfs ) ) st = uv0;

   // Get the colour for the pixel.

   vec4 colour = texture2D ( iChannel0, fract(st) );

   // Optionaly make any tears transparent.

   if ( any ( edge_tears ) ) colour = vec4(0.0);

   return colour;

 }

// ============================================================================
// == The shader's main routine ===============================================
// ============================================================================

void main ( void )
 {
   // Get the normalised coordinates of the current point.

   vec2 uv = Emu_Normalise_to_Window ( gl_FragCoord.xy );

   // Apply the ripple and update the shader's outputs.

   gl_FragColor = Emu_Ripple_XY(uv) * gl_Color;

 }

// ============================================================================
